# -*- coding: utf-8 -*-
import collections


Permission = collections.namedtuple('Permission', 'id name inherited value')


class Acl(object):

    """Access Control List functionality for managing users' roles and
    permissions.

    By default, the user model contains an `acl` attribute, which allows
    access to the Acl object.

    Attributes:
        allow_default (boolean): Whether or not to allow/deny access if the
                                 permission has not been set on that role.

    """
    allow_default = True
    _permissions = None

    def __init__(self, user):
        """Initializes the Acl.

        Args:
            watson.auth.models.UserMixin user: The user to validate against
        """
        self.user = user

    @property
    def permissions(self):
        if not self._permissions:
            self._generate_user_permissions()
        return self._permissions

    def has_role(self, role_key):
        """Validates a role against the associated roles on a user.

        Args:
            role_key (string|tuple|list): The role(s) to validate against.
        """
        for role in self.user.roles:
            if isinstance(role_key, (list, tuple)) and role.key in role_key:
                return True
            elif role.key == role_key:
                return True
        return False

    def has_permission(self, permission):
        """Check to see if a user has a specific permission.

        If the permission has not been set, then it access will be granted
        based on the allow_default attribute.

        Args:
            permission (string): The permission to find.
        """
        if permission not in self.permissions.keys():
            return self.allow_default
        if not self.permissions[permission].value:
            return False
        return True

    def _generate_user_permissions(self):
        """Internal method to generate the permissions for the user.

        Retrieve all the permissions associated with the users roles, and then
        merge the users individual permissions to overwrite the inherited
        role permissions.
        """
        permissions = {}
        for role in self.user.roles:
            permissions.update(
                {permission.permission.key: Permission(
                    id=permission.permission_id,
                    name=permission.permission.name,
                    inherited=1,
                    value=permission.value)
                    for permission in role.permissions})
        permissions.update(
            {permission.permission.key: Permission(
                id=permission.permission_id,
                name=permission.permission.name,
                inherited=0, value=permission.value)
                for permission in self.user.permissions})
        self._permissions = permissions



# -*- coding: utf-8 -*-
from watson.framework.debug import abc

data = []


class User(abc.Panel):
    title = 'Auth'
    icon = 'user'

    @property
    def request(self):
        return self.event.params['context']['request']

    @property
    def user(self):
        user = None
        if hasattr(self.request, 'user') and getattr(self.request, 'user'):
            user = getattr(
                getattr(self.request, 'user'),
                self.application.config['auth']['model']['columns']['username']
            )
        return user

    def render(self):
        return self._render({
            'request': self.request,
            'user': self.user
        })

    def render_key_stat(self):
        return self.user




# -*- coding: utf-8 -*-
import bcrypt
import uuid
from watson.auth import models
from watson.common import imports, decorators
from watson.db.contextmanagers import transaction_scope
from sqlalchemy.orm import exc


def generate_password(password, rounds=10, encoding='utf-8'):
    """Generate a new password based on a random salt.

    Args:
        password (string): The password to generate the hash off
        rounds (int): The complexity of the hashing

    Returns:
        mixed: The generated password and the salt used
    """
    salt = bcrypt.gensalt(rounds)
    return bcrypt.hashpw(password.encode(encoding), salt), salt


def check_password(password, existing_password, salt, encoding='utf-8'):
    """Validate a password against an existing password and the salt used to
    generate it.

    Args:
        password (string): The password to validate
        existing_password (string): The password to validate against
        salt (string): The salt used to generate the existing_password

    Returns:
        boolean: True/False if valid or invalid
    """
    return bcrypt.hashpw(password.encode(encoding), salt) == existing_password


def generate_forgotten_token():
    """Generates a unique identifier that can be used to validate a forgotten
    password request.
    """
    return uuid.uuid4().hex


class Authenticator(object):
    """Authenticates a user against the database.

    Attributes:
        session: The SqlAlchemy session to use.
        user_model (string): The class name of the user model.
        user_id_field (string): The field that is to be used as the username.
    """
    session = None
    user_model = None
    user_id_field = None
    _config = None

    def __init__(self, config, session, user_model, user_id_field):
        self._config = config
        self.session = session
        self.user_model = user_model
        self.user_id_field = user_id_field

    @decorators.cached_property
    def user_model_class(self):
        """The of the user_model to be used.
        """
        return imports.load_definition_from_string(self.user_model)

    @property
    def user_query(self):
        return self.session.query(self.user_model_class)

    def get_user(self, username):
        """Retrieves a user from the database based on their username.

        Args:
            username (string): The username of the user to find.
        """
        user_field = getattr(self.user_model_class, self.user_id_field)
        try:
            user = self.user_query.filter(user_field == username).one()
        except exc.NoResultFound:
            return None
        return user

    def get_user_by_id(self, id):
        """Retrieves a user from the database based on their id.

        Args:
            id (int): The id of the user to find.
        """
        try:
            user = self.user_query.filter_by(id=id).one()
        except exc.NoResultFound:
            return None
        return user

    def authenticate(self, username, password):
        """Validate a user against a supplied username and password.

        Args:
            username (string): The username of the user.
            password (string): The password of the user.
        """
        password_config = self._config['password']
        if len(password) > password_config['max_length']:
            return None
        user = self.get_user(username)
        if user:
            if check_password(password, user.password, user.salt,
                              password_config['encoding']):
                return user
        return None

    def assign_user_to_session(self, user, request, session_key):
        """Assign a user to the session.
        """
        request.user = user
        request.session[session_key] = getattr(
            user, self.user_id_field)


class ForgottenPasswordTokenManager(object):
    """Manages a users forgotten password.

    Attributes:
        email_field (string): The email property of the user model.
        mailer (watson.mail.backends.abc.Base): The mailer backend used to send the email.
    """
    email_field = None
    mailer = None
    _config = None

    def __init__(self, config, session, mailer, email_field):
        self._config = config
        self.session = session
        self.mailer = mailer
        self.email_field = email_field

    def create_token(self, user, request):
        """Create a new forgotten password token.

        Email the user their reset password link. The template for this email
        can be overridden via `emails/forgotten-password.html`.

        Args:
            user (watson.auth.models.User): The user who forgot their password
            request (watson.http.messages.Request): The HTTP request
        """
        token = models.ForgottenPasswordToken(token=generate_forgotten_token())
        user.forgotten_password_tokens.append(token)
        with transaction_scope(self.session) as session:
            session.add(user)
        self.mailer.send(
            subject=self._config['subject_line'],
            from_=self._config['from'],
            template='emails/forgotten-password',
            body={'user': user, 'token': token, 'request': request},
            to=getattr(user, self.email_field))
        return token

    def delete_token(self, token):
        """Delete a token.

        Args:
            token: The token to be deleted
        """
        with transaction_scope(self.session) as session:
            session.delete(token)

    def get_token(self, token, model=models.ForgottenPasswordToken):
        """Retrieve a user based on the supplied token.

        Args:
            token (string): The forgotten password token
        """
        return self.session.query(
            model).filter_by(
                token=token).order_by(model.id.desc()).first()

    def update_user_password(self, token, password):
        """Update the users password.

        Once the user has been updated, make sure that the token has been
        deleted to prevent further access of that token.
        """
        token.user.password = password
        with transaction_scope(self.session) as session:
            session.add(token.user)
            session.delete(token)



# -*- coding: utf-8 -*-
from watson.validators import abc


class Match(abc.Validator):
    """Provides the ability to ensure two fields have the same value.
    """

    def __init__(self, field, message='Supplied value for "{field}" does not match.'):
        """Initial the validator.

        Args:
            field (str): The field name to match
        """
        self.message = message
        self.field = field

    def __call__(self, value, form):
        matching_field_value = getattr(form, self.field)
        if value != matching_field_value:
            raise ValueError(self.message.format(
                field=self.field))
        return True



# -*- coding: utf-8 -*-
from watson import form
from watson.form import fields
from watson.auth import validators


class Login(form.Form):
    """A standard login form containing username and password fields.
    """
    username = fields.Text(required=True, label='Username')
    password = fields.Password(required=True, label='Password')
    submit = fields.Submit(button_mode=True, label='Login')


class ForgottenPassword(form.Form):
    """A standard forgotten password form.
    """
    username = fields.Text(required=True, label='Username')
    submit = fields.Submit(
        button_mode=True,
        label='Request Reset')


class ResetPassword(form.Form):
    """A standard reset password form.
    """
    password = fields.Password(
        required=True,
        validators=[validators.Match(field='confirm_password')],
        label='Password')
    confirm_password = fields.Password(required=True, label='Confirm Password')
    submit = fields.Submit(button_mode=True, label='Reset')



# -*- coding: utf-8 -*-
from watson.console import command, ConsoleError
from watson.console.decorators import arg
from watson.common import imports
from watson.db.session import NAME as session_name
from watson.db.contextmanagers import transaction_scope
from watson.di import ContainerAware
from watson.auth import authentication


class BaseAuthCommand(ContainerAware):
    __ioc_definition__ = {
        'init': {
            'config': lambda container: container.get('application.config')['auth']
        }
    }

    def __init__(self, config):
        self.config = config


def ensure_session_in_container(container, session):
    if not session:
        session = 'default'
    try:
        return container.get(session_name.format(session))
    except:
        raise ConsoleError('No database named {}'.format(session))


class Auth(command.Base, BaseAuthCommand):
    """Authentication commands.
    """
    name = 'auth'

    @arg('password')
    def generate_password(self, password):
        """Generates a password based on the application settings.
        """
        salt, new_password = authentication.generate_password(password)
        self.write('Generated hashed password from "{}"'.format(password))
        self.write('Password: {}'.format(new_password))
        self.write('Salt: {}'.format(salt))

    @arg('database', optional=True)
    def list_roles(self, database):
        """Lists all available roles.

        Args:
            database: The name of the database session.
        """
        session = ensure_session_in_container(self.container, database)
        from watson.auth.models import Role
        for role in session.query(Role):
            self.write('Role: {} (key: {})'.format(role.name, role.key))

    @arg('key')
    @arg('name')
    @arg('database', optional=True)
    def add_role(self, key, name, database):
        """Adds a new role to the database.

        Args:
            key: The identifier for the role
            name: The human readable name for the role
            database: The name of the database session.
        """
        session = ensure_session_in_container(self.container, database)
        from watson.auth.models import Role
        role = Role(key=key, name=name)
        with transaction_scope(session) as session:
            session.add(role)
            self.write('Role: {} (key: {}) added!'.format(name, key))

    @arg('key')
    @arg('name')
    @arg('database', optional=True)
    def add_permission(self, key, name, database):
        """Adds a new permission to the database.

        Args:
            key: The identifier for the permission
            name: The human readable name for the permission
            database: The name of the database session.
        """
        session = ensure_session_in_container(self.container, database)
        from watson.auth.models import Permission
        permission = Permission(key=key, name=name)
        with transaction_scope(session) as session:
            session.add(permission)
            self.write('Permission: {} (key: {}) added!'.format(name, key))

    @arg('permission_key')
    @arg('role_key')
    @arg('value', optional=True)
    @arg('database', optional=True)
    def add_permission_to_role(self, permission_key, role_key, value, database):
        """Attaches a permission to a role, with a default of False.

        Args:
            permission_key: The identifier for the permission
            role_key: The identifier for the permission
            value: The value of the permission (1 or 0)
            database: The name of the database session.
        """
        session = ensure_session_in_container(self.container, database)
        from watson.auth.models import Role
        from watson.auth.models import Permission
        permission = session.query(Permission).filter_by(key=permission_key).first()
        role = session.query(Role).filter_by(key=role_key).first()
        enabled = True if value else False
        role.add_permission(permission, value=enabled)
        session.commit()
        self.write(
            'Added permission {} ({}) to role {} ({}) with value: {}'.format(
                permission.name, permission.key,
                role.name, role.key,
                enabled))

    @arg('username')
    @arg('role_key')
    @arg('database', optional=True)
    def add_role_to_user(self, username, role_key, database):
        session = ensure_session_in_container(self.container, database)
        auth_authenticator = self.container.get('auth_authenticator')
        user = auth_authenticator.get_user(username)
        from watson.auth.models import Role
        role = session.query(Role).filter_by(key=role_key).first()
        user.roles.append(role)
        session.commit()
        self.write(
            'Added role {} ({}) to user {} ({})'.format(
                role.name, role.key,
                getattr(user, self.config['model']['columns']['username']),
                user.id))

    @arg('username')
    @arg('permission_key')
    @arg('value', optional=True)
    @arg('database', optional=True)
    def add_permission_to_user(self, username, permission_key, value, database):
        session = ensure_session_in_container(self.container, database)
        auth_authenticator = self.container.get('auth_authenticator')
        user = auth_authenticator.get_user(username)
        from watson.auth.models import Permission
        permission = session.query(Permission).filter_by(key=permission_key).first()
        enabled = True if value else False
        user.add_permission(permission, value=enabled)
        session.commit()
        self.write(
            'Added permission {} ({}) to user {} ({}) with value: {}'.format(
                permission.name, permission.key,
                getattr(user, self.config['model']['columns']['username']),
                user.id, enabled))

    @arg('username')
    @arg('password')
    @arg('database', optional=True)
    def create_user(self, username, password, database):
        """Create a new user.

        Args:
            username: The username of the user
            password: The password of the user
            database: The name of the database session.
        """
        session = ensure_session_in_container(self.container, database)
        db_config = self.config['model']
        user_model = db_config['class']
        model_class = imports.load_definition_from_string(user_model)
        with transaction_scope(session) as session:
            user = model_class()
            setattr(user, db_config['columns']['username'], username)
            setattr(user, 'password', password)
            session.add(user)
            self.write('Created user {}'.format(username))

    @arg('database', optional=True)
    def list_permissions(self, database):
        """Lists all available permissions.

        Args:
            database: The name of the database session.
        """
        session = ensure_session_in_container(self.container, database)
        from watson.auth.models import Permission
        for permission in session.query(Permission):
            self.write('Permission: {} (key: {})'.format(
                permission.name, permission.key))



# -*- coding: utf-8 -*-
from watson.common import datastructures
from watson.console.command import find_commands_in_module
from watson.di import ContainerAware
from watson.framework import events
from watson.auth import config, commands


class Init(ContainerAware):

    """Bootstraps watson.auth into the event system of watson.
    """

    __ioc_definition__ = {
        'property': {
            'app_config': 'application.config',
        }
    }

    app_config = None

    def __call__(self, event):
        app = event.target
        self.setup_database(event)
        self.setup_config()
        self.load_default_commands(app.config)
        self.setup_listeners()
        self.setup_authenticator()

    def setup_database(self, event):
        # Initialize watson.db if it hasn't been added to the app config
        db_listener = ('watson.db.listeners.Init', 1, False)
        if db_listener not in self.app_config['events'][events.INIT]:
            listener = self.container.get('watson.db.listeners.Init')
            listener(event)

    def load_default_commands(self, config):
        """Load some existing
        """
        existing_commands = config.get('commands', [])
        db_commands = find_commands_in_module(commands)
        db_commands.extend(existing_commands)
        config['commands'] = db_commands

    def setup_config(self):
        auth_config = datastructures.dict_deep_update(
            config.defaults, self.app_config.get('auth', {}))
        self.app_config['auth'] = auth_config
        self.container.get('jinja2_renderer').add_package_loader(
            'watson.auth', 'views')

    def setup_authenticator(self):
        for name, definition in config.definitions.items():
            dependency_config = datastructures.dict_deep_update(
                config.definitions[name],
                self.app_config['dependencies']['definitions'].get(name, {}))
            self.app_config['dependencies']['definitions'][name] = dependency_config
            self.container.definitions[name] = dependency_config

    def setup_listeners(self):
        dispatcher = self.container.get('shared_event_dispatcher')
        dispatcher.add(
            events.ROUTE_MATCH,
            self.container.get('watson.auth.listeners.Route'),
            1,
            False)


class Route(ContainerAware):

    """Listens for a route event and attempts to inject the user into the
    request if one has been authenticated.
    """

    def __call__(self, event):
        auth_config = self.container.get('application').config['auth']
        request = event.params['context']['request']
        request.user = None
        user_id = request.session[auth_config['session']['key']]
        if user_id:
            authenticator = self.container.get('auth_authenticator')
            request.user = authenticator.get_user(user_id)



# -*- coding: utf-8 -*-

defaults = {
    'model': {
        'columns': {
            'username': 'username',
            'email': 'email'
        }
    },
    'authenticator': {
        'session': 'default',
        'password': {
            'max_length': 30,
            'encoding': 'utf-8'
        },
        'urls': {
            'unauthenticated': 'login',
            'unauthorized': 'unauthorized',
        }
    },
    'login': {
        'redirect_to_unauthenticated': False,
        'urls': {
            'success': '/',
            'route': 'login'
        },
        'form': {
            'class': 'watson.auth.forms.Login',
            'username_field': 'username',
            'password': 'password',
            'messages': {
                'invalid': 'Invalid username and/or password.'
            }
        }
    },
    'logout': {
        'urls': {
            'success': '/',
            'route': 'logout'
        }
    },
    'forgotten_password': {
        'urls': {
            'route': 'forgotten-password'
        },
        'subject_line': 'A password reset request has been made',
        'form': {
            'class': 'watson.auth.forms.ForgottenPassword',
            'username_field': 'username',
            'messages': {
                'success': 'A password reset request has been sent to your email.',
                'invalid': 'Could not find your account in the system, please try again.',
            }
        }
    },
    'reset_password': {
        'authenticate_on_reset': False,
        'urls': {
            'route': 'reset-password',
            'success': '/',
            'invalid': '/'
        },
        'form': {
            'class': 'watson.auth.forms.ResetPassword',
            'username_field': 'username',
            'messages': {
                'success': 'Your password has been changed successfully.',
                'invalid': 'Could not find your account in the system, please try again.',
            }
        }
    },
    'session': {
        'key': 'watson.user'
    },
}


definitions = {
    'auth_authenticator': {
        'item': 'watson.auth.authentication.Authenticator',
        'init': {
            'config': lambda container: container.get('application.config')['auth']['authenticator'],
            'session': lambda container: container.get('sqlalchemy_session_{0}'.format(container.get('application.config')['auth']['authenticator']['session'])),
            'user_model': lambda container: container.get('application.config')['auth']['model']['class'],
            'user_id_field': lambda container: container.get('application.config')['auth']['model']['columns']['username']
        },
    },
    'auth_forgotten_password_token_manager': {
        'item': 'watson.auth.authentication.ForgottenPasswordTokenManager',
        'init': {
            'config': lambda container: container.get('application.config')['auth']['forgotten_password'],
            'session': lambda container: container.get('sqlalchemy_session_{0}'.format(container.get('application.config')['auth']['authenticator']['session'])),
            'email_field': lambda container: container.get('application.config')['auth']['model']['columns']['email'],
            'mailer': lambda container: container.get('mailer')
        },
    }
}



# -*- coding: utf-8 -*-
from datetime import datetime
from sqlalchemy import (Column, Integer, String, DateTime, ForeignKey,
                        SmallInteger)
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relationship
from watson.common import imports
from watson.auth import authentication, authorization
from watson.db.models import Model
from watson.db.utils import _table_attr


class Permission(Model):
    id = Column(Integer, primary_key=True)
    name = Column(String(255))
    key = Column(String(255))
    created_date = Column(DateTime, default=datetime.now)

    def __repr__(self):
        return '<{0} key:{1} name:{2}>'.format(
            imports.get_qualified_name(self), self.key, self.name)


class Role(Model):
    id = Column(Integer, primary_key=True)
    name = Column(String(255))
    key = Column(String(255))
    permissions = relationship('RolesHasPermission',
                               backref='roles')
    created_date = Column(DateTime, default=datetime.now)

    def add_permission(self, permission, value=1):
        """Adds a permission to the role.

        Args:
            Permission permission: The permission to attach
            int value: The value to give the permission, can be either:
                        0 - deny
                        1 - allow
        """
        role_permission = RolesHasPermission(value=value)
        role_permission.permission = permission
        self.permissions.append(role_permission)

    def __repr__(self):
        return '<{0} key:{1} name:{2}>'.format(
            imports.get_qualified_name(self), self.key, self.name)


class UserMixin(object):

    """Common user fields, custom user classes should extend this as well as
    Model.

    Attributes:
        string id_field: The name of the field to use as the id for the user

    Columns:
        string _password: The password of the user, aliased by self.password
        string salt: The salt used to generate the password
        list roles: The roles associated with the user
        list permissions: The permissions associated with the user, overrides
                          the permissions associated with the role.
        date created_date: The time the user was created.
        date updated_date: The time the user was updated.
    """
    __tablename__ = 'users'
    _acl_class = authorization.Acl
    _acl = None
    id = Column(Integer, primary_key=True)
    _password = Column(String(255), name='password')
    salt = Column(String(255), nullable=False)
    created_date = Column(DateTime, default=datetime.now)
    updated_date = Column(DateTime, default=datetime.now)

    @property
    def acl(self):
        """Convenience method to access the users ACL.

        See watson.auth.authorization.Acl for more information.
        """
        if not self._acl:
            self._acl = self._acl_class(self)
        return self._acl

    @declared_attr
    def permissions(cls):
        return relationship(UsersHasPermission, backref='user', cascade='all')

    @declared_attr
    def roles(cls):
        return relationship(Role,
                            secondary=UsersHasRole.__tablename__,
                            backref='roles', cascade=None)

    @declared_attr
    def forgotten_password_tokens(cls):
        return relationship(ForgottenPasswordToken, backref='user', cascade='all')

    @property
    def password(self):
        """Return the password.
        """
        return self._password

    @password.setter
    def password(self, password):
        """Automatically generates the hashed password and salt when set.

        Args:
            string password: The password to set.
        """
        _pass, salt = authentication.generate_password(password)
        self._password = _pass
        self.salt = salt

    def touch(self):
        """Updates the date the user was modified.
        """
        self.updated_date = datetime.now()

    def add_permission(self, permission, value=1):
        """Adds a permission to the user.

        This overrides any permission given by the associated roles.

        Args:
            Permission permission: The permission to attach
            int value: The value to give the permission, can be either:
                        0 - deny
                        1 - allow
        """
        user_permission = UsersHasPermission(value=value)
        user_permission.permission = permission
        self.permissions.append(user_permission)

    def __repr__(self):
        return '<{0} id:{1}>'.format(imports.get_qualified_name(self), self.id)


class RolesHasPermission(Model):
    role_id = Column(Integer,
                     ForeignKey(_table_attr(Role, 'id')),
                     primary_key=True)
    permission_id = Column(Integer,
                           ForeignKey(_table_attr(Permission, 'id')),
                           primary_key=True)
    permission = relationship(Permission)
    value = Column(SmallInteger, default=0)
    created_date = Column(DateTime, default=datetime.now)


class UsersHasPermission(Model):
    user_id = Column(Integer,
                     ForeignKey(_table_attr(UserMixin, 'id')),
                     primary_key=True)
    permission_id = Column(Integer,
                           ForeignKey(_table_attr(Permission, 'id')),
                           primary_key=True)
    permission = relationship(Permission)
    value = Column(SmallInteger, default=0)
    created_date = Column(DateTime, default=datetime.now)


class UsersHasRole(Model):
    user_id = Column(Integer,
                     ForeignKey(_table_attr(UserMixin, 'id')),
                     primary_key=True)
    role_id = Column(Integer,
                     ForeignKey(_table_attr(Role, 'id')),
                     primary_key=True)


class ForgottenPasswordToken(Model):
    id = Column(Integer, primary_key=True)
    token = Column(String(255))
    user_id = Column(Integer,
                     ForeignKey(_table_attr(UserMixin, 'id')))
    created_date = Column(DateTime, default=datetime.now)

    def __repr__(self):
        return '<{0} user id:{1}>'.format(
            imports.get_qualified_name(self), self.user.id)



# -*- coding: utf-8 -*-
from urllib import parse
from watson.common import imports
from watson.framework import exceptions


def auth(func=None, roles=None, permissions=None, requires=None,
         unauthenticated_url=None, unauthorized_url=None,
         should_404=False, redirect=False):
    """Makes a controller action require an authenticated user.

    By setting additional roles and permissions, finer control can be given
    to the resource.

    Args:
        func (callable): the function that is being wrapped
        roles (string|iterable): The roles that the user must have
        permissions (string|iterable): The permissions the user must have
        requires (iterable): An iterable of watson.validator.Validator objects
                             that allows for custom validations outside of
                             standard roles/permissions.
        unauthenticated_url (string): The url to redirect to if the user
                                      is not logged in.
        unauthorized_url (string): The url to redirect to if the user does
                                   not have permission.
        should_404 (boolean): Raise a 404 instead of redirecting.
        redirect (boolean): Remember the url to redirect to after login.

    Returns:
        The controller response.

    Example:

    .. code-block:: python

        class MyController(controllers.Action):
            def index_action(self):
                return 'Index'

            @auth(roles='admin', permissions='view')
            def protected_action(self):
                return 'Authenticated users only'
    """
    def decorator(func):
        def wrapper(self, *args, **kwargs):
            auth_config = self.container.get('application.config')['auth']
            user_id = self.request.session[auth_config['session']['key']]
            config_urls = auth_config['authenticator']['urls']
            unauthent_url = unauthenticated_url or config_urls['unauthenticated']
            # User has not been authenticated
            if not user_id:
                if redirect or auth_config['login']['redirect_to_unauthenticated']:
                    unauthent_url = '{}?redirect={}'.format(
                        unauthent_url, parse.quote_plus(str(self.request.url)))
                return self.redirect(unauthent_url)
            authenticator = self.container.get('auth_authenticator')
            if not hasattr(self.request, 'user'):
                setattr(self.request, 'user', authenticator.get_user(user_id))
            user = self.request.user
            if not user:
                return self.redirect(unauthent_url)

            # User has been authenticated, but is not authorized
            unauthor_url = unauthorized_url or config_urls['unauthorized']
            no_role = roles and not user.acl.has_role(roles)
            no_permission = permissions and not user.acl.has_permission(
                permissions)
            no_require = False
            if requires:
                for require in requires:
                    if not require(self.request.user):
                        no_require = True
                        break
            if no_role or no_permission or no_require:
                if should_404:
                    raise exceptions.NotFoundError('Not found')
                return self.redirect(unauthor_url)

            # User is valid
            return func(self, **kwargs)
        return wrapper
    if func:
        return decorator(func)
    else:
        return decorator


def login(func=None, method='POST', form_class=None, auto_redirect=True):
    """Attempts to authenticate a user if the required fields have been posted.

    By setting auto_redirect to False, the user roles and permissions can be
    checked within the login route and redirected from there.

    Args:
        func (callable): the function that is being wrapped
        method (string): The HTTP method that authentication will be performed
                         against.
        form_class (string): The qualified class name of the form.
        auto_redirect (boolean): Whether or not to automatically redirect to a
                                 different url on successful login.

    Example:

    .. code-block:: python

        class MyController(controllers.Action):
        @login(auto_redirect=False)
        def login_action(self):
            return 'Logged In'
    """
    def decorator(func):
        def wrapper(self, *args, **kwargs):
            if not hasattr(self.request, 'user'):
                self.request.user = None
            auth_config = self.container.get('application.config')['auth']
            login_config = auth_config['login']
            valid = True
            if self.request.user:
                # User is already authenticated, redirect to success url
                if auto_redirect:
                    return self.redirect(login_config['urls']['success'])
            form_config = login_config['form']
            form_class_module = form_class or form_config['class']
            form = imports.load_definition_from_string(form_class_module)(
                action=self.request)
            kwargs['form'] = form
            if self.request.is_method(method):
                # Form has been posted, validate the user
                form.data = self.request
                if form.is_valid():
                    authenticator = self.container.get('auth_authenticator')
                    username_field = getattr(form, form_config['username_field'])
                    password_field = getattr(form, form_config['password'])
                    user = authenticator.authenticate(
                        username=username_field,
                        password=password_field)
                    if user:
                        authenticator.assign_user_to_session(
                            user,
                            self.request,
                            auth_config['session']['key'])
                        if auto_redirect:
                            redirect_url = login_config['urls']['success']
                            if self.request.get['redirect']:
                                redirect_url = parse.unquote_plus(
                                    self.request.get['redirect'])
                            return self.redirect(redirect_url, clear=True)
                    else:
                        valid = False
                else:
                    valid = False
            if not valid:
                self.flash_messages.add(
                    form_config['messages']['invalid'], 'error')
                return self.redirect(login_config['urls']['route'])
            return func(self, **kwargs)
        return wrapper
    if func:
        return decorator(func)
    else:
        return decorator  # pragma: no cover


def logout(func=None, redirect_url=None):
    """Attempts to log a user out of the application.

    Args:
        redirect_url (string): The url to redirect to.

    Example:

    .. code-block:: python

        class MyController(controllers.Action):
            def index_action(self):
                return 'Index'

            @logout(redirect_url='home')
            def logout_action(self):
                pass
    """
    def decorator(func):
        def wrapper(self, *args, **kwargs):
            auth_config = self.container.get('application').config['auth']
            logout_config = auth_config['logout']
            del self.request.session[auth_config['session']['key']]
            redirect_to = logout_config['urls']['success']
            if redirect_url:
                redirect_to = redirect_url
            func(self, **kwargs)
            self.request.user = None
            return self.redirect(redirect_to)
        return wrapper
    if func:
        return decorator(func)
    else:
        return decorator


def forgotten(func=None, method='POST', form_class=None):
    """Finds a user and sends them a reset password request email.

    Args:
        func (callable): the function that is being wrapped.
        method (string): The HTTP method that request must match.
        form_class (string): The qualified class name of the form.
    """
    def decorator(func):
        def wrapper(self, *args, **kwargs):
            auth_config = self.container.get('application').config['auth']
            forgotten_config = auth_config['forgotten_password']
            if hasattr(self.request, 'user') and self.request.user:
                # User is already authenticated, redirect to success url
                return self.redirect(auth_config['login']['urls']['success'])
            form_config = forgotten_config['form']
            form_class_module = form_class or form_config['class']
            form = imports.load_definition_from_string(form_class_module)(
                action=self.request)
            kwargs['form'] = form
            if self.request.is_method(method):
                form.data = self.request
                invalid = True
                if form.is_valid():
                    authenticator = self.container.get('auth_authenticator')
                    forgotten_password_token_manager = self.container.get(
                        'auth_forgotten_password_token_manager')
                    username_field = getattr(form, form_config['username_field'])
                    user = authenticator.get_user(username_field)
                    if user:
                        forgotten_password_token_manager.create_token(
                            user, self.request)
                        invalid = False
                        self.flash_messages.add(
                            form_config['messages']['success'], 'error')
                        return self.redirect(str(self.request.url))
                if invalid:
                    self.flash_messages.add(
                        form_config['messages']['invalid'], 'error')
            return func(self, **kwargs)
        return wrapper
    if func:
        return decorator(func)
    else:
        return decorator


def reset(func=None, method='POST', form_class=None, authenticate_on_reset=False):
    """Resets a users password if the token matches.

    If a token is not matched the user is redirected to the specified route/url.

    Args:
        func (callable): the function that is being wrapped.
        method (string): The HTTP method that request must match.
        form_class (string): The qualified class name of the form.
        authenticate_on_reset (bool): Automatically log the user in on success.
    """
    def decorator(func):
        def wrapper(self, *args, **kwargs):
            auth_config = self.container.get('application').config['auth']
            request = self.request
            if hasattr(request, 'user') and request.user:
                # User is already authenticated, redirect to success url
                return self.redirect(auth_config['login']['urls']['success'])
            reset_config = auth_config['reset_password']
            form_config = reset_config['form']
            form_class_module = form_class or form_config['class']
            form = imports.load_definition_from_string(form_class_module)(
                action=request)
            kwargs['form'] = form
            forgotten_password_token_manager = self.container.get(
                'auth_forgotten_password_token_manager')
            token = forgotten_password_token_manager.get_token(
                request.get['token'])
            if request.is_method(method) and token:
                authenticator = self.container.get('auth_authenticator')
                form.data = request
                if form.is_valid():
                    self.flash_messages.add(
                        form_config['messages']['success'], 'error')
                    redirect_url = reset_config['urls']['success']
                    auto_login = authenticate_on_reset or reset_config['authenticate_on_reset']
                    if auto_login:
                        authenticator.assign_user_to_session(
                            token.user,
                            request,
                            auth_config['session']['key'])
                        forgotten_password_token_manager.update_user_password(
                            token, form.password)
                        redirect_url = auth_config['login']['urls']['success']
                    forgotten_password_token_manager.delete_token(token)
                    func(self, **kwargs)
                    return self.redirect(redirect_url)
            if not token:
                self.flash_messages.add(
                    form_config['messages']['invalid'], 'error')
                return self.redirect(reset_config['urls']['invalid'])
            return func(self, **kwargs)
        return wrapper
    if func:
        return decorator(func)
    else:
        return decorator
